/*
  Lithium 0.1

  Copyright (C) 2001 Teti Paolo.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  (LICENSE file) along with this program; if not, write to
  the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/

package lithium2;

import java.util.*;
import java.rmi.*;
import java.rmi.server.*;


/**
 *
 * LithiumServer: the remote skeletons evaluator
 *
 *
 *@author Paolo Teti
 *@version 0.1
 */
public class LithiumServer extends UnicastRemoteObject
    implements LithiumInterface
{

    private Vector remoteServerWorker = null; // Skeletons Table
    private int numOfSkeletons = 0;
    private Stack returnPoints = null;
    private static final String VERSION = "0.1";
    private Stack loopContest = null;
    private int nTasks = 1;

    public LithiumServer () throws RemoteException
    {
	returnPoints = new Stack();
	returnPoints.ensureCapacity(LithiumConst.NMEDOFLOOPS);
	loopContest = new Stack();
	loopContest.ensureCapacity(LithiumConst.NMEDOFLOOPS);
    }



    public String getVersion()
    {
	return VERSION;
    }

    public void setRemoteWorker(Vector skeletonsList)
    {
	remoteServerWorker = skeletonsList;
	numOfSkeletons = skeletonsList.size();
    }


    public TaskItem[] execute(TaskItem[] task)
    {
	boolean returns = false;
	TaskItem[] ret = null;

	System.out.println("Esecuzione task : "+nTasks++);

	while (!returns)
	    {
		Object info = ((InfoSkel)remoteServerWorker.
			       elementAt(task[0].nextSkeleton)).info;

		int typeOfSkel =((InfoSkel)remoteServerWorker.
				 elementAt(task[0].nextSkeleton)).skelType;

		int jump = ((Integer)((InfoSkel)remoteServerWorker.
				      elementAt(task[0].nextSkeleton)).incomplete.
			    firstElement()).intValue();

		Object[] extraInfo = ((InfoSkel)remoteServerWorker.
				      elementAt(task[0].nextSkeleton)).extraInfo;

		int address = ((InfoSkel)remoteServerWorker.
			       elementAt(task[0].nextSkeleton)).next;


		switch (typeOfSkel)
		    {
			case SkeletonsCode.SEQ:
			{
			    //  System.out.println("EXEC SEQ!!!!\n");
			    // Execute the SEQ skeleton.
			    ret = new TaskItem[1];
			    task[0].dataTask=((JSkeletons)info).run(task[0].dataTask);

			    ret[0] = task[0];

			    if (address==numOfSkeletons-1)
				task[0].nextSkeleton = LithiumConst.END;
			    else task[0].nextSkeleton = address;

			    //  task[0].nextSkeleton = LithiumConst.END;

			    returns = true;
			    break;
			}

		    case SkeletonsCode.DIVIDE:
			{
			    Object[] tmp = ((Divide)info).run(task[0].dataTask);
			    ret = new TaskItem[ret.length];
			    for (int j=0;j<ret.length;j++)
				{
				    ret[j].dataTask = tmp[j];
				    ret[j].nextSkeleton = task[0].nextSkeleton;

				    ret[j].toXVP = new int[task[0].toXVP.length];
				    System.arraycopy(task[0].toXVP,0,ret[j].toXVP,0,task[0].toXVP.length);
				    ret[j].toYVP = new int[ret[0].toYVP.length];
				    System.arraycopy(task[0].toYVP,0,ret[j].toYVP,0,task[0].toXVP.length);
				    ret[j].globalIndex = task[0].globalIndex;
				    System.arraycopy(task[0].loopCounter,0,ret[j].loopCounter,0,
						     task[0].loopCounter.length);
				}

			    returns = true;
			    break;
			}

		    case SkeletonsCode.RECOMP:
			{
			    Object[] o = new Object[task.length];
			    for (int i=0;i<task.length;i++)
				o[i] = task[i].dataTask;
			    Object val = ((Recomp)info).run(o);
			    ret = new TaskItem[1];
			    ret[0].dataTask = val;
			    ret[0].toXVP = new int[task[0].toXVP.length];
			    System.arraycopy(task[0].toXVP,0,ret[0].toXVP,0,task[0].toXVP.length);
			    ret[0].toYVP = new int[ret[0].toYVP.length];
			    System.arraycopy(task[0].toYVP,0,ret[0].toYVP,0,task[0].toXVP.length);
			    ret[0].globalIndex = task[0].globalIndex;
			    System.arraycopy(task[0].loopCounter,0,ret[0].loopCounter,0,
					     task[0].loopCounter.length);

			    if (!returnPoints.empty())
				ret[0].nextSkeleton = ((Integer)returnPoints.pop()).intValue();
			    else
				ret[0].nextSkeleton = address;
			    returns = true;
			    break;
			}

		    case SkeletonsCode.WHILE:
			{
			    if (!((Condition)info).apply(task[0].dataTask))
				task[0].nextSkeleton =jump;
			    else
				task[0].nextSkeleton = address;
			    break;
			}


		    case SkeletonsCode.IF:
			{
			    if (((Condition)info).apply(task[0].dataTask))
				task[0].nextSkeleton = address;
			    else
				task[0].nextSkeleton = jump;
			    break;
			}


		    case SkeletonsCode.LOOP:
			{
			    int numIter = ((Integer)info).intValue();
			    int val;

			    if (task[0].loopCounter[task[0].lastLC] == LithiumConst.UNDEF)
				{
				    task[0].loopCounter[task[0].lastLC++] = numIter;
				    val = numIter;
				}
			    else
				val = task[0].loopCounter[task[0].lastLC];
			    if (val>0)
				{
				    val--;
				    task[0].loopCounter[task[0].lastLC]--;
				}
			    else
				{
				    task[0].loopCounter[task[0].lastLC]= LithiumConst.UNDEF;
				    task[0].lastLC--;
				}

			    task[0].nextSkeleton = address;
			    break;
			}


		    case SkeletonsCode.MAP:
			{
			    MapUtils eval = new MapUtils(extraInfo);

			    if (task[0].toXVP[task[0].lastVP] == LithiumConst.UNDEF &&
				task[0].toYVP[task[0].lastVP] == LithiumConst.UNDEF)
				{
				    int numOfRule = extraInfo.length;
				    int dimY = ((Integer)extraInfo[numOfRule-1]).intValue();
				    int dimX = ((Integer)extraInfo[numOfRule-2]).intValue();

				    int [] vpDim = eval.getVP();
				    dimX = vpDim[0];
				    dimY = vpDim[1];

				    Object[][] o =
					eval.applyRule((Object[])task[0].dataTask);

				    ret = new TaskItem[dimX*dimY];
				    int m=0;
				    for (int i=0;i<dimX;i++)
					for (int j=0;j<dimY;j++)
					    {
						TaskItem tmpTask = new TaskItem();
						tmpTask.dataTask =
						    ((Vector)o[i][j]).toArray();
						tmpTask.lastVP = task[0].lastVP+1;

						tmpTask.toXVP = new int[task[0].toXVP.length];
						System.arraycopy(task[0].toXVP,0,tmpTask.toXVP,0,task[0].toXVP.length);
						tmpTask.toYVP = new int[task[0].toYVP.length];
						System.arraycopy(task[0].toYVP,0,tmpTask.toYVP,0,task[0].toXVP.length);

						tmpTask.toXVP[tmpTask.lastVP] = i;
						tmpTask.toYVP[tmpTask.lastVP] = j;
						tmpTask.globalIndex = task[0].globalIndex;
						tmpTask.nextSkeleton = address;
						System.arraycopy(task[0].loopCounter,0,tmpTask.loopCounter,0,
								 task[0].loopCounter.length);

						ret[m] = tmpTask;
						m++;
					    }
				}
			    else
				{
				    eval.recomp(task);
				    ret = new TaskItem[1];
				    ret[0].dataTask = eval.getResult();

				    ret[0].globalIndex = task[0].globalIndex;
				    ret[0].lastVP = task[0].lastVP;

				    ret[0].toXVP = new int[task[0].toXVP.length];
				    System.arraycopy(task[0].toXVP,0,ret[0].toXVP,0,task[0].toXVP.length);
				    ret[0].toYVP = new int[ret[0].toYVP.length];
				    System.arraycopy(task[0].toYVP,0,ret[0].toYVP,0,task[0].toXVP.length);

				    ret[0].toXVP[ret[0].lastVP] = LithiumConst.UNDEF;
				    ret[0].toYVP[ret[0].lastVP] = LithiumConst.UNDEF;
				    ret[0].lastVP--;
				    ret[0].nextSkeleton = address;
				    System.arraycopy(task[0].loopCounter,0,ret[0].loopCounter,0,
						     task[0].loopCounter.length);
				}

			    returns = true;
			    break;

			}

		    case SkeletonsCode.DC:
			{
			    if (((Condition)info).apply(task[0].dataTask))
				// The DC Worker
				task[0].nextSkeleton=task[0].nextSkeleton+2;
			    else
				{
				    // Divide Rule
				    task[0].nextSkeleton++;

				    returnPoints.push(new Integer(task[0].nextSkeleton+1));
				}
			}

		    }
	    }
	return (ret);
    }



    private static void printUsage()
    {
	System.out.println("\n-security  Install a Security RMI Manager.\n");
	System.exit(0);
    }


    private static void printVersionInfo()
    {
	System.out.println("\nLithium " + VERSION +
			   " Copyrigth 2001 - Paolo Teti");
	System.out.println("------------------------------------------");
	System.out.println("JRE Version:          " +
			   System.getProperty("java.version"));
	System.out.println("OS Information:       " +
			   System.getProperty("os.name") + " " +
			   System.getProperty("os.version") + " " +
			   System.getProperty("os.arch"));
	System.out.println("User Login:           " +
			   System.getProperty("user.name"));
	System.out.println("------------------------------------------\n");
    }


    public static void main(String [] args) throws Exception
    {
	if (System.getProperty("java.version").compareTo("1.2")<0)
	    {
		System.err.println("!!!WARNING: Use JDK version 1.2 or higher!!!");
		System.exit(-1);
	    }

	if (args.length!=0 )
	    if (args[0].equals("-security"))
		System.setSecurityManager(new RMISecurityManager());
	    else
		printUsage();

	printVersionInfo();

	LithiumServer worker = new LithiumServer();
	Naming.rebind("LithiumWorker",worker);
    }


} // LithiumServer





