/**
 * LithiumTaskPool.java
 *
 *
 *
 *
 * @author Paolo Teti
 * @version 0.1
 */

package lithium2;
import java.util.*;


class LithiumTaskPool
{
    private int typeOfTaskPool = LithiumConst.LIFO;
    private LinkedList /*taskpool,*/ respool;
    private LinkedList[] taskpool;
    private int[] paramNum;
    private int totalSize;
    private boolean stop = false;

    // Create the Task and Result Pool
    LithiumTaskPool(int type,Object[] paramNum)
    {
	typeOfTaskPool = type;

	respool  = new LinkedList();
	this.paramNum = new int[paramNum.length];
	taskpool = new LinkedList[paramNum.length];
	for (int i=0;i<paramNum.length;i++)
	    {
		taskpool[i] = new LinkedList();
		this.paramNum[i] = ((Integer)paramNum[i]).intValue();
	    }
    }


    private int findFired()
    {
	int j;
	for (j=0;j<taskpool.length;j++)
	    if (taskpool[j].size()>=paramNum[j])
		break;
	return j;
    }

    // Fine sequenza
    synchronized void stop()
    {
	stop = true;
    }

    // usato dai control Thread per prelevare i token
    synchronized TaskItem[] getTask() throws Exception
    {
	int find = findFired();
	TaskItem[] o = new TaskItem[paramNum[find]];

       	switch (typeOfTaskPool)
	    {
	    case LithiumConst.LIFO:
		{
		    for (int i=0;i<paramNum[find];i++)
			{
			    //System.out.println("Letto dato ..\n");
			    o[i] =(TaskItem) (taskpool[find].getLast());
			    //   System.out.println("---> " +
			    //	       ((Integer)((TaskItem)o[i]).
			    //		dataTask).
			    //		intValue());
			    taskpool[find].removeLast();
			}
		    return o;
		}
	    case LithiumConst.FIFO:
		{
		    for (int i=0;i<paramNum[find];i++)
			{
			    o[i] = (TaskItem)(taskpool[find].getLast());
			    taskpool[find].removeFirst();
			}
		    return o;
		}
	    }
	totalSize = totalSize-paramNum[find];

	return (null);
    }


    // metodo di utlita' per la ricerca
    synchronized private Object search(int ind)
    {
	Iterator i = respool.iterator();
	while (true)
	    if (((TaskItem)i.next()).globalIndex == ind) break;
	    else i.next();

	return i;
    }


    //
    // Prelava il task dal respool conm numero d'ordine ind
    //
    synchronized Object getTaskFromResPool(int ind) throws Exception
    {
	switch (typeOfTaskPool)
	    {
	    case LithiumConst.LIFO:
		if (ind==LithiumConst.UNDEF)
		    {
			Object o = respool.getLast();
			respool.removeLast();
			return o;
		    }
		else  return search(ind);

	    case LithiumConst.FIFO:
		if (ind==LithiumConst.UNDEF)
		    {
			Object o = respool.getFirst();
			respool.removeFirst();
			return o;
		    }
		else return search(ind);
	    }
	return (null);
    }



    synchronized void addTaskPool(TaskItem tasks) throws Exception
    {

	switch (typeOfTaskPool)
	    {
	    case LithiumConst.LIFO:
		{
		    taskpool[tasks.nextSkeleton].addLast(tasks);
		    break;
		}
	    case LithiumConst.FIFO:
		{
		    taskpool[tasks.nextSkeleton].addFirst(tasks);
		    break;
		}
	    }
	totalSize++;
    }

    // usato dai thread per aggiungere un task
    synchronized void addTask(TaskItem[] tasks) throws Exception
    {
	switch (typeOfTaskPool)
	    {
		case LithiumConst.LIFO:
		{
		    for (int i=0;i<tasks.length;i++)
			{
			    ResData res = new ResData();
			    res.dataTask = tasks[i].dataTask;
			    res.globalIndex = tasks[i].globalIndex;

			    if (tasks[i].nextSkeleton==(LithiumConst.END))
			    {
				// System.out.println("Realmente nel res\n");
				respool.addFirst(res);
			    }
			    else
				taskpool[tasks[i].nextSkeleton].
				    addFirst(res);
			}
			    break;
		}

	    case LithiumConst.FIFO:
		{
		    for (int i=0;i<tasks.length;i++)
			{

			    ResData res = new ResData();
			    res.dataTask = tasks[i].dataTask;
			    res.globalIndex = tasks[i].globalIndex;

			    if (tasks[i].nextSkeleton==(LithiumConst.END))
				respool.addLast(res);
			    else
				taskpool[tasks[i].nextSkeleton].
				    addLast(res);
			}
			    break;
		}
	    }
	totalSize++;
    }


    synchronized boolean isEmpty() throws Exception
    {
	return (totalSize==0 || stop);
    }

    synchronized boolean isResEmpty() throws Exception
    {
	return (respool.size()==0);
    }

    synchronized public int taskPoolSize() throws Exception
    {
	return (totalSize);
    }

} // LithiumTaskPool

