/*
  Lithium 0.1

  Copyright (C) 2001 Teti Paolo.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  (LICENSE file) along with this program; if not, write to
  the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/


package lithium2;

import java.rmi.*;
import java.util.*;
import java.lang.*;


/**
 *
 *<b>Implements the parallel evaluator of
 * a generic skeletons composition. </b>
 *
 *
 * @author Paolo Teti
 * <A HREF="mailto:tetipaol@cli.di.unipi.it"> tetipaol@cli.di.unipi.it</A>
 * @version 0.1 */

public class Ske
{

    private String[] hostList = null;
    private int nPE = 0;
    private int pos = -1;
    private boolean outOfOrder = true;
    private Thread [] controlThreadList = null;
    static LithiumInterface [] slave = null;
    private long finishTime = 0;
    private long startTime = 0;
    private long startUpTime = 0;
    private int i = 0;
    private int streamLength = 0;
    private int positionInStream = 0;
    private boolean normalForm = true;
    private boolean createTaskPool  =  true;
    static Vector skeletonsTree = null;
    LithiumTaskPool taskPool;
    private int schedulingRule = LithiumConst.LIFO;
    private int globalPosition = 0;
    private boolean secondTime =  false;
    private Vector paramNum;
    private int numberOfLoop;
    private int numberOfMap;
    private Stack returnsPoint;
    private boolean stop = false;
    /**
     * Create a parallel evaluator of a generic
     * skeletons compositions.
     *
     */
    public Ske()
    {
	// Create the skeletons Tree
	skeletonsTree = new Vector();
	// Init the stack
	returnsPoint = new Stack();
        paramNum = new Vector();
    }

    /**
     * Show the version of the parallel evaluetor
     *
     * @return a String that rappresent the version
     */
    public String getVersion()
    {
	return LithiumVersion.VERSION;
    }

    /**
     * Test the compatibility of the server version
     * with this client version of Lithium
     * @return <code>true<code> this version of the
     * client is compatibile, <>false<> otherwise.
     */
    public boolean controlServerCompatibility() throws Exception
    {
	for (int i=0;i<hostList.length;i++)
	    {
		LithiumInterface server = (LithiumInterface)
		    Naming.lookup("rmi://"+hostList[i]+"/LithiumWorker");
		if (server.getVersion().compareTo(LithiumVersion.VERSION)!=0)
		    return false;
	    }
	return true;
    }


    /**
     * Setup the set of hosts to use for
     * the parallel evaluation of the skeletons
     * composition.
     *
     * @param host Is a list of hosts names
     *
     */
    public void addHosts(String[] host)
    {
	nPE = host.length;
	hostList = new String[nPE];
	System.arraycopy(host,0,hostList,0,nPE);
	slave = new LithiumInterface[nPE];
	controlThreadList = new Thread[nPE];
    }

    /**
     * Get the actual set of hosts.
     *
     * @return A String vector contains the actual set of hosts
     */
    public String[] getHostsList()
    {
	return hostList;
    }

    /**
     * Get the system configuration from an XML file.
     * <b> Note: This method is actually incomplete. </b>
     * @param name the XML Configurazion file
     */
    public void getConfigurationFromFile(String name)
    {
	LithiumXMLParser parser = new LithiumXMLParser(name);
    }

    /**
     * Declare the end of the imput stream.
     */
    public void stopStream()
    {
	stop = true;
    }


    /**
     * Declare the use of the LIFO Scheduling Rule for the tasks
     *
     */
    public void setLifoSchedulingRule()
    {
	schedulingRule = LithiumConst.LIFO;
    }

    /**
     * Declare the use of the FIFO Scheduling Rule for the tasks
     *
     */
    public void setFifoSchedulingRule()
    {
	schedulingRule = LithiumConst.FIFO;
    }

    /**
     *
     * Not preserve the input-output order. This is the default
     * rule used.
     */
    public void outOfOrderResults()
    {
	outOfOrder = true;
    }

    /**
     *
     * Preserve the input-output order.
     */
    public void inOrderResult()
    {
	outOfOrder = false;
    }

    /**
     *
     * Declare the use of the normal form rewriting rule
     */
    public void setOptimizations()
    {
	normalForm = true;
    }

    /**
     *
     * Declare the non-use of the normal form rewriting rule
     */
    public void resetOptimizations()
    {
	normalForm = false;
    }


    public int taskPoolSize() throws Exception
    {
	return taskPool.taskPoolSize();
    }

    /**
     * Setup the task pool.
     *
     * @param tasks The task to be computed
     */
    public void setupTaskPool(Object tasks) throws Exception
    {
	TaskItem dataIn =  new TaskItem();

	dataIn.dataTask = tasks;
	dataIn.nextSkeleton = 0;
	if (!stop)
	    {
		if (createTaskPool)
		    {
			positionInStream = 0;

			taskPool = new
			    LithiumTaskPool(schedulingRule, paramNum.toArray());

			createTaskPool = false;
		    }

		dataIn.globalIndex = positionInStream;

   		dataIn.toXVP = new int[numberOfMap+1];
		dataIn.toYVP = new int[numberOfMap+1];
		dataIn.toXVP[0] = LithiumConst.UNDEF;
		dataIn.toYVP[0] = LithiumConst.UNDEF;

		dataIn.loopCounter = new int[numberOfLoop+1];
		dataIn.loopCounter[0] = LithiumConst.UNDEF;
		dataIn.lastLC = 0;
		dataIn.lastVP = 0;
		positionInStream++;

		taskPool.addTaskPool(dataIn);

		}
	else
	    if (!createTaskPool)
		taskPool.stop();
    }


    /**
     * Read the task pool
     *
     * @return The task computed.
     */
    public Object readTaskPool() throws Exception
    {
       	if (outOfOrder)
	    return (((ResData)taskPool.
		     getTaskFromResPool(LithiumConst.UNDEF)).dataTask);
	else
	    {
		pos++;
		return (((TaskItem)taskPool.
			 getTaskFromResPool(pos)).dataTask);
	    }
    }

    /**
     * Test if the Results Pool is empty
     *
     * @return true if the Results Pool is empty
     */
    public boolean isResEmpty() throws Exception
    {
	return (taskPool.isResEmpty());
    }


    /**
     * Evaluate the time spent to execute a skeletons
     * composition.
     *
     * @return The Elapsed time, in milliseconds, of the parallel application
     */
    public long getElapsedTime()
    {
	return (finishTime);
    }

    /**
     * Evaluate the time spent to execute a skeletons
     * composition.
     *
     * @return The Elapsed time, in seconds, of the parallel application
     */
    public long getElapsedTimeInSec()
    {
	return (finishTime/1000);
    }


    /**
     * Evaluate the time spent to inizialize the runtime support
     *
     * @return The Startup time, in milliseconds, of the runtime support
     */
    public long getStartUpTime()
    {
	return (startUpTime);
    }


    /**
     * Evaluate the time spent to inizialize the runtime support
     *
     * @return The Startup time, in seconds, of the runtime support
     */
    public long getStartUpTimeInSec()
    {
	return (startUpTime/1000);
    }


    // Costruzione dell'albero DF degli skeletons
    private void getMacroDataFlowGraph(JSkeletons[] skel,int padre)
    {
	int instanceTest = LithiumConst.UNDEF;

	for (int ind=0;ind<skel.length;ind++)
	    {
		if (skel[ind].getClass().isInstance(new While(null,null)))
		    instanceTest = SkeletonsCode.WHILE;

		if (skel[ind].getClass().isInstance(new Loop(null,0)))
		    instanceTest = SkeletonsCode.LOOP;

		if (skel[ind].getClass().isInstance(new Farm(null)))
		    instanceTest = SkeletonsCode.FARM;

		if (skel[ind].getClass().isInstance(new Pipeline()))
		    instanceTest = SkeletonsCode.PIPE;

		if (skel[ind].getClass().isInstance(new If(null,null,null)))
		    instanceTest = SkeletonsCode.IF;

		if (skel[ind].getClass().isInstance(new Map(null)))
		    instanceTest = SkeletonsCode.MAP;

		if (skel[ind].getClass().isInstance(new
		    DivideConquer(null,null,null,null)))
		    instanceTest = SkeletonsCode.DC;

		switch(instanceTest)
		    {

		    case SkeletonsCode.LOOP:
			{
			    numberOfLoop++;
			    globalPosition++;
			    JSkeletons[] saveTmp = new JSkeletons[1];
			    saveTmp[0] = (JSkeletons)(skel[ind].
						      getSkeletonInfo()[1]);

			    InfoSkel tmp = new InfoSkel();
			    tmp.incomplete = new Vector();
			    tmp.skelType = instanceTest;
			    tmp.padreType = padre;
			    tmp.info=((skel[ind].getSkeletonInfo())[0]);
			    Integer incValue = new Integer(globalPosition);
			    tmp.incomplete.add(incValue);
			    returnsPoint.push(new Integer(globalPosition));
			    paramNum.add(new Integer(1));
			    skeletonsTree.add(tmp);

			    getMacroDataFlowGraph(saveTmp,tmp.padreType);
			    break;
			}

		    case SkeletonsCode.IF:
			{
			    globalPosition++;
			    JSkeletons[] saveTmpSx = new JSkeletons[1];
			    saveTmpSx[0] = (JSkeletons)skel[ind].
				getSkeletonInfo()[1];

			    JSkeletons[] saveTmpDx = new JSkeletons[1];

			    saveTmpDx[1] = (JSkeletons)skel[ind].
				getSkeletonInfo()[2];
			    paramNum.add(new Integer(1));
			    InfoSkel tmp = new InfoSkel();
			    tmp.incomplete = new Vector();
			    tmp.skelType = instanceTest;
			    tmp.padreType = padre;
			    tmp.info=((skel[ind].getSkeletonInfo())[0]);

			    skeletonsTree.add(tmp);
			    instanceTest = LithiumConst.UNDEF;

			    // SX Path
			    getMacroDataFlowGraph(saveTmpSx,tmp.padreType);

			    Integer incValue = new Integer(globalPosition);
			    tmp.incomplete.add(incValue);
			    skeletonsTree.set(skeletonsTree.indexOf(tmp),tmp);

			    int savePosition = globalPosition;

			    // DX Path
			    getMacroDataFlowGraph(saveTmpDx,tmp.padreType);

			    InfoSkel save = (InfoSkel)skeletonsTree.elementAt(savePosition);
			    save.next = globalPosition;
			    skeletonsTree.set(savePosition,save);

			    break;
			}

		    case SkeletonsCode.WHILE:
			{
			    globalPosition++;
			    JSkeletons[] saveTmp = new JSkeletons[1];
			    saveTmp[0] = (JSkeletons)skel[ind].
				getSkeletonInfo()[1];

			    InfoSkel tmp = new InfoSkel();
			    paramNum.add(new Integer(1));
			    tmp.incomplete = new Vector();
			    tmp.skelType = instanceTest;
			    tmp.padreType = padre;
			    tmp.info=((skel[ind].getSkeletonInfo())[0]);

			    skeletonsTree.add(tmp);

			    // Memorizzazione del punto di ritorno
			    returnsPoint.push(new Integer(globalPosition));
			    getMacroDataFlowGraph(saveTmp,tmp.padreType);

			    Integer incValue = new Integer(globalPosition);
			    int cont = skeletonsTree.indexOf(tmp);
			    tmp.incomplete.add(incValue);
			    skeletonsTree.set(cont,tmp);

			    break;
			}

		    case SkeletonsCode.DC:
			{
			    globalPosition++;
			    InfoSkel tmp = new InfoSkel();
			    tmp.padreType = padre;
			    tmp.skelType = instanceTest;
			    tmp.incomplete = new Vector();


			    // The base-condition
			    tmp.info = skel[ind].getSkeletonInfo()[0];

			    skeletonsTree.add(tmp);
			    JSkeletons[] saveTmp = new JSkeletons[1];
			    saveTmp[0] = (JSkeletons)skel[ind].
				getSkeletonInfo()[1];
			    paramNum.add(new Integer(1));

			    InfoSkel tmp1 = new InfoSkel();
			    tmp1.skelType = SkeletonsCode.DIVIDE;

			    globalPosition++;
			    saveTmp[0] = (JSkeletons)skel[ind].
				getSkeletonInfo()[2];
			    tmp1.info = saveTmp[0];
			    tmp1.next = globalPosition-1;

			    skeletonsTree.add(tmp1);

			    globalPosition++;
			    InfoSkel tmp2 = new InfoSkel();
			    tmp2.skelType = SkeletonsCode.DIVIDE;
			     paramNum.add(new Integer(1));
			    tmp.incomplete.add(new Integer(globalPosition));

			    saveTmp[0] = (JSkeletons)skel[ind].
				getSkeletonInfo()[3];
			    tmp2.info = saveTmp[0];
			    skeletonsTree.add(tmp2);
			    paramNum.add(new Integer(((Recomp)tmp2.info).size()));
			    returnsPoint.push(new Integer(globalPosition));
		    // Worker
			    getMacroDataFlowGraph(saveTmp,instanceTest);

			    tmp2.next = globalPosition;
			    skeletonsTree.set(skeletonsTree.indexOf(tmp2),tmp2);

			    break;
			}

		    case SkeletonsCode.MAP:
			{
			    numberOfMap++;
			    globalPosition++;
			    InfoSkel tmp = new InfoSkel();
			    int packDim = skel[ind].getSkeletonInfo().length;
			    tmp.info = null;
			    tmp.extraInfo = new Object[packDim-4];
			    for (i=0;i<packDim-4;i++)
				tmp.extraInfo[i] = skel[ind].getSkeletonInfo()[i];
			    tmp.skelType = instanceTest;
			    tmp.padreType = padre;
			    tmp.incomplete = null;
			    // Il seguito e' la regola di ricomposizone.
			    tmp.next = globalPosition+1;
                            skeletonsTree.add(tmp);
			    // Numero di token per abilitare il nodo: 1
			    paramNum.add(new Integer(1));

			    // Nodo ricomposizione
			    InfoSkel tmp2 = new InfoSkel();
			    tmp2.skelType = instanceTest;
			    tmp2.padreType = padre;
			    tmp2.incomplete = null;
			    tmp2.extraInfo = new Object[3];
			    for (int i=0;i<packDim-4;i++)
				tmp2.extraInfo[i] = skel[ind].getSkeletonInfo()[i];

			    MapUtils eval = new MapUtils(tmp2.extraInfo);

			    paramNum.add(new Integer(eval.getVP()[0]*eval.getVP()[1]));
			    skeletonsTree.add(tmp2);
			    returnsPoint.push(new Integer(globalPosition));

			    JSkeletons[] saveTmp = new JSkeletons[1];
			    saveTmp[0] = (JSkeletons)skel[ind].
				getSkeletonInfo()[packDim-1];

			    getMacroDataFlowGraph(saveTmp,instanceTest);

			    break;
			}

		    case SkeletonsCode.FARM:
			{
			    if (skel[ind].getSkeletonInfo()!=null)
				getMacroDataFlowGraph((JSkeletons[])
						      skel[ind].
						      getSkeletonInfo(),
						      SkeletonsCode.FARM);
			    break;
			}

		    case SkeletonsCode.PIPE:
			{
			    System.out.println("Trovato PIPE\n");
			    if (skel[ind].getSkeletonInfo()!=null)
				{
				    // System.out.println("Dentro IF PIPE\n");
				    //int len = skel[ind].getSkeletonInfo().length;
				    //JSkeletons [] tmpSkel = new JSkeletons[len];

				    getMacroDataFlowGraph((JSkeletons[])
							  skel[ind].
							  getSkeletonInfo(),
							  SkeletonsCode.PIPE);
				}
			    System.out.println("Fine Trovato PIPE\n");
			    break;
			}

		    default: // Sequential Skeleton
			{
			    InfoSkel tmp = new InfoSkel();
			    tmp.incomplete = new Vector();
			    tmp.incomplete.add(new Integer(0));
			    tmp.padreType = padre;
			    tmp.skelType = SkeletonsCode.SEQ;
			    tmp.info = skel[ind];

			    if (!returnsPoint.empty())
				tmp.next = ((Integer)returnsPoint.pop()).intValue();
			    else tmp.next = globalPosition;
			    skeletonsTree.add(tmp);
			    paramNum.add(new Integer(1));
			    globalPosition++;

			    break;
			}

		    }// switch
	    }
    }




    //
    // Create the Normal-Form skeletons tree given a
    // generic Macro Data-Flow Graph ...
    //
    private void normalizeSkeletonsTree()
    {
	int j = 0;
	int position = 0;
	InfoSkel actualGenericElement = null;
	InfoSkel nextGenericElement = null;
	boolean find = false;
	boolean end = false;
	Vector incRetPoint = new Vector();

	while (j<(skeletonsTree.size()-1))
	    {
		actualGenericElement = (InfoSkel)skeletonsTree.elementAt(j);
		nextGenericElement = (InfoSkel)skeletonsTree.elementAt(j+1);

		if (actualGenericElement == null && nextGenericElement == null)
		    {
			if (actualGenericElement.skelType == SkeletonsCode.WHILE ||
			    actualGenericElement.skelType == SkeletonsCode.LOOP ||
			    actualGenericElement.skelType == SkeletonsCode.IF ||
			    actualGenericElement.skelType == SkeletonsCode.DC)
			    incRetPoint.add(new Integer(j));

			if ((actualGenericElement.padreType == SkeletonsCode.FARM ||
			     actualGenericElement.padreType == SkeletonsCode.DC ||
			     actualGenericElement.padreType == SkeletonsCode.PIPE ||
			     actualGenericElement.skelType == SkeletonsCode.SEQ))
			    {
				position=(j+1);
				while ((nextGenericElement.padreType == SkeletonsCode.PIPE ||
					nextGenericElement.padreType == SkeletonsCode.FARM ||
					actualGenericElement.padreType == SkeletonsCode.DC ||
					nextGenericElement.skelType == SkeletonsCode.SEQ)
				       && position<(skeletonsTree.size()-1))
				    {
					end = (position<skeletonsTree.size());
					position++;
					nextGenericElement = (InfoSkel)skeletonsTree.
					    elementAt(position);
				    }

				int dragdrop = (position-j);

				if (dragdrop>1)
				    {
					Pipeline collaps = new Pipeline();
					for (int i=0;i<dragdrop;i++)
					    collaps.addWorker(((JSkeletons)
							       ((InfoSkel)
								skeletonsTree.elementAt(j+i)).info));

					for (int i=0;i<dragdrop-i; i++)
					    skeletonsTree.remove(j+i);

					actualGenericElement.skelType = SkeletonsCode.SEQ;
					actualGenericElement.padreType = SkeletonsCode.PIPE;
					actualGenericElement.info = collaps;
					skeletonsTree.set(j,actualGenericElement);

					// Recomp the returns point
					while (!incRetPoint.isEmpty())
					    {
						int pos = ((Integer)incRetPoint.firstElement()).
						    intValue();
						incRetPoint.remove(0);

						InfoSkel tmpInfo =
						    (InfoSkel)skeletonsTree.elementAt(pos);
						tmpInfo.incomplete.removeAllElements();
						tmpInfo.incomplete.add(new Integer(pos-dragdrop));
						skeletonsTree.add(tmpInfo);
					    }

				    }
			    }
		    }
		j++;
	    }
    }

    /**
     * Setup the skeletons composition to evaluate.
     * @param sk a generic skeletons composition
     */
    public void skelProgram(JSkeletons sk)
    {
	JSkeletons[] sk2 = new JSkeletons[1];
	sk2[0] = sk;

	getMacroDataFlowGraph(sk2,LithiumConst.ROOT);
	if (normalForm) // Create the Normal Form ....
	    normalizeSkeletonsTree();
    }


    /**
     * Parallel evaluation of the skeletons composition.
     *
     * @exception BadHostException when the hostlist is incorrect.
     */
    public void parDo() throws Exception
    {
	if (hostList==null) throw new
	    BadHostException(LithiumExc.BADHOST);

	// Create a set of Control Thread to implement
	// the macro data-flow run-time support.
	startUpTime = System.currentTimeMillis();

	for(int i=0;i<nPE;i++)
	    {
		controlThreadList[i] = new
		    ControlThread(i,hostList[i],skeletonsTree,taskPool);
		// Max Priority...
		controlThreadList[i].setPriority(Thread.MAX_PRIORITY);
	    }

	System.out.println(" Thread creati ...\n");
	// Calcolo tempo di startup dei Thread di controllo.
	startUpTime = (System.currentTimeMillis()-startUpTime);

	startTime = System.currentTimeMillis();

	// Start all the Control Threads
	for(int i=0;i<nPE;i++)
	    controlThreadList[i].start();


	// Join threads
	for(int i=0;i<nPE;i++)
	    controlThreadList[i].join();

	// Calcolo del tempo di completamento senza startup
       	finishTime =  (System.currentTimeMillis()-startTime);
    }

} // Ske





