/*
  Lithium 0.1

  Copyright (C) 2001 Teti Paolo.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  (LICENSE file) along with this program; if not, write to
  the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/

/**
 * While.java
 *
 * The While skeleton
 *
 *
 * @author Paolo Teti
 * @version 0.1
 */

package lithium2;

public class While extends JSkeletons
{

    JSkeletons workerWhile = null;
    Condition condition = null;


    /**
     * <b> Create the While skeleton </b>
     *
     * @param workerWhile Skeleton to be executed
     * @param condition The condition to be used
     */
    public While(JSkeletons workerWhile,
		 Condition cond)
    {
	this.workerWhile = workerWhile;
	this.condition   = cond;
    }


    Object[] getSkeletonInfo()
    {
	Object[] info = new Object[2];
	info[0] = condition;
	info[1] = workerWhile;
     	return (info);
    }


    /**
     * Execute the skeleton in sequenzial mode.
     * (i.e like a sequential skeleton)
     *
     * @param task The task to compute.
     * @return The result of the computation.
     */
    public Object run(Object task)
    {
	Object whileTask = task;

	while (condition.apply(whileTask))
	    whileTask = workerWhile.run(whileTask);

	return (whileTask);
    }



} // While





